import math
from abc import ABC

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.enumerations.tokenisation_flags import TokenisationFlags
from scoda.exceptions.tokenisation_exception import TokenisationException
from scoda.sequences.sequence import Sequence
from scoda.settings.settings import PPQN
from scoda.tokenisation.base_tokenisation import BaseTokeniser


class BaseTransposedNotelikeTokeniser(BaseTokeniser, ABC):

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__()

        self.flags[TokenisationFlags.RUNNING_TIME_SIG] = running_time_sig


class TransposedNotelikeTokeniser(BaseTransposedNotelikeTokeniser):
    """Tokeniser that uses transposed temporal representation with a note-like approach, i.e., all occurrences of a note
    are shown first before any other note is handled. Note that input sequences are expected to represent bars,
    as the transposed representation is done on a per-bar basis.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [        4] ... wait
    [        5] ... play
    [  6 -  29] ... value definition
    [ 30 - 117] ... note
    [118 - 132] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 133

    def __init__(self, running_value: bool, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

        self.flags[TokenisationFlags.RUNNING_VALUE] = running_value

    def tokenise(self, bar_sequence: Sequence, apply_buffer: bool = True, reset_time: bool = True,
                 insert_border_tokens: bool = False) -> list[int]:
        tokens = []
        event_pairings = bar_sequence.abs.get_message_time_pairings(
            [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE, MessageType.INTERNAL])

        event_pairings_by_note = dict()

        # Entries consist of (message type, hashable)
        ordered_type_value_pairs = list()

        for event_pairing in event_pairings:
            message = event_pairing[0]

            if message.message_type == MessageType.NOTE_ON:
                if not any(x[1] == message.note for x in ordered_type_value_pairs):
                    ordered_type_value_pairs.append((MessageType.NOTE_ON, message.note))

                event_pairings_by_note.setdefault(message.note, list())
                event_pairings_by_note[message.note].append(event_pairing)
            elif message.message_type == MessageType.TIME_SIGNATURE:
                if any(x[1] == MessageType.TIME_SIGNATURE for x in ordered_type_value_pairs):
                    raise TokenisationException("Bar contains more than one time signature")

                ordered_type_value_pairs.insert(0, (MessageType.TIME_SIGNATURE, MessageType.TIME_SIGNATURE))

                event_pairings_by_note.setdefault(MessageType.TIME_SIGNATURE, list())
                event_pairings_by_note[MessageType.TIME_SIGNATURE].append(event_pairing)
            elif message.message_type == MessageType.INTERNAL:
                if any(x[1] == MessageType.INTERNAL for x in ordered_type_value_pairs):
                    raise TokenisationException("Bar contains more than one internal message")

                ordered_type_value_pairs.append((MessageType.INTERNAL, MessageType.INTERNAL))

                event_pairings_by_note.setdefault(MessageType.INTERNAL, list())
                event_pairings_by_note[MessageType.INTERNAL].append(event_pairing)

        for ordered_type_value_pair in ordered_type_value_pairs:
            msg_type = ordered_type_value_pair[0]
            event_pairings_for_key = event_pairings_by_note[ordered_type_value_pair[1]]

            # Define note
            if msg_type == MessageType.NOTE_ON:
                msg_note = event_pairings_for_key[0][0].note
                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                tokens.append(msg_note - 21 + 30)

            for event_pairing in event_pairings_for_key:
                msg_time = event_pairing[0].time

                if msg_type == MessageType.NOTE_ON:
                    msg_note = event_pairing[0].note
                    msg_value = event_pairing[1].time - msg_time

                    if not (21 <= msg_note <= 108):
                        raise TokenisationException(f"Invalid note: {msg_note}")

                    # Check if message occurs at current time, if not place rest messages
                    if not self.cur_time == msg_time:
                        self.cur_rest_buffer += msg_time - self.cur_time
                        tokens.extend(
                            self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4,
                                                                      index_time_def=6))

                    # Check if value of note has to be defined
                    if not (self.prv_value == msg_value and self.flags.get(TokenisationFlags.RUNNING_VALUE, False)):
                        tokens.extend(self._general_tokenise_flush_time_buffer(msg_value, index_time_def=6))

                    # Play note
                    tokens.append(5)

                    self.cur_time_target = max(self.cur_time_target, self.cur_time + msg_value)
                    self.prv_type = MessageType.NOTE_ON
                    self.prv_value = msg_value
                elif msg_type == MessageType.TIME_SIGNATURE:
                    msg_numerator = event_pairing[0].numerator
                    msg_denominator = event_pairing[0].denominator

                    numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                    # Check if time signature has to be defined
                    if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG,
                                                                               False)):
                        self.cur_rest_buffer += msg_time - self.cur_time
                        tokens.extend(
                            self._notelike_tokenise_flush_rest_buffer(apply_target=False, wait_token=4,
                                                                      index_time_def=6))
                        tokens.append(numerator - 2 + 118)

                    self.prv_type = MessageType.TIME_SIGNATURE
                    self.prv_numerator = numerator
                elif msg_type == MessageType.INTERNAL:
                    self.cur_rest_buffer += msg_time - self.cur_time
                    self.prv_type = MessageType.INTERNAL

            # Reset time to beginning of bar
            self.cur_time = 0

        if apply_buffer:
            tokens.extend(
                self._notelike_tokenise_flush_rest_buffer(apply_target=True, wait_token=4, index_time_def=6))

        if reset_time:
            self.reset_time()

        # Insert bar border mark
        tokens.append(3)

        if insert_border_tokens:
            tokens.insert(0, 1)
            tokens.append(2)

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        time_bar_start = 0

        cur_time_bar = 0

        prv_type = None
        prv_note = math.nan
        prv_value = math.nan
        prv_numerator = math.nan

        for token in tokens:
            if token <= 2:
                prv_type = "sequence_control"
            elif token == 3:
                time_bar_start += prv_numerator * PPQN / 2
                cur_time_bar = 0

                prv_type = "bar_border"
            elif token == 4:
                cur_time_bar += prv_value

                prv_type = MessageType.WAIT
            elif token == 5:
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_ON, note=prv_note,
                            time=time_bar_start + cur_time_bar))
                seq.add_absolute_message(
                    Message(message_type=MessageType.NOTE_OFF, note=prv_note,
                            time=time_bar_start + cur_time_bar + prv_value))

                prv_type = MessageType.NOTE_ON
            elif 6 <= token <= 29:
                if prv_type == "value_definition":
                    prv_value += token - 5
                else:
                    prv_value = token - 5

                prv_type = "value_definition"
            elif 30 <= token <= 117:
                cur_time_bar = 0

                prv_note = token - 30 + 21
                prv_type = "note_definition"
            elif 118 <= token <= 132:
                numerator = token - 118 + 2

                seq.add_absolute_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, time=time_bar_start + cur_time_bar,
                            numerator=numerator, denominator=8)
                )

                prv_type = MessageType.TIME_SIGNATURE
                prv_numerator = numerator
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq



from __future__ import annotations

import enum


class TokeniserType(enum.Enum):
    STANDARD_MIDILIKE_TOKENISER = "standard_midilike_tokeniser"
    RELATIVE_MIDILIKE_TOKENISER = "relative_midilike_tokeniser"
    COF_MIDILIKE_TOKENISER = "cof_midilike_tokeniser"
    STANDARD_NOTELIKE_TOKENISER = "standard_notelike_tokeniser"
    LARGE_VOCABULARY_NOTELIKE_TOKENISER = "large_vocabulary_notelike_tokeniser"
    RELATIVE_NOTELIKE_TOKENISER = "relative_notelike_tokeniser"
    COF_NOTELIKE_TOKENISER = "cof_notelike_tokeniser"
    LARGE_VOCABULARY_COF_NOTELIKE_TOKENISER = "large_vocabulary_cof_notelike_tokeniser"
    GRIDLIKE_TOKENISER = "gridlike_tokeniser"
    TRANSPOSED_NOTELIKE_TOKENISER = "transposed_notelike_tokeniser"



from __future__ import annotations

from scoda.enumerations.message_type import MessageType
from scoda.misc.music_theory import Key


class Message:
    """Class representing a musical message.
    """

    def __init__(self, message_type: MessageType = None, note: int = None, velocity: int = None, control: int = None,
                 numerator: int = None, denominator: int = None, key: Key = None, time: int = None,
                 program: int = None) -> None:
        super().__init__()
        self.message_type = message_type
        self.time = time
        self.note = note
        self.velocity = velocity
        self.control = control
        self.program = program
        self.numerator = numerator
        self.denominator = denominator
        self.key = key

    def __copy__(self) -> Message:
        return Message(message_type=self.message_type, note=self.note, velocity=self.velocity, control=self.control,
                       numerator=self.numerator, denominator=self.denominator, key=self.key, time=self.time,
                       program=self.program)

    # def __eq__(self, o: object) -> bool:
    #     if not isinstance(o, Message):
    #         return False
    #
    #     return (
    #             self.message_type == o.message_type and
    #             self.note == o.note and
    #             self.velocity == o.velocity and
    #             self.control == o.control and
    #             self.numerator == o.numerator and
    #             self.denominator == o.denominator and
    #             self.key == o.key and
    #             self.time == o.time and
    #             self.program == o.program
    #     )

    # def __hash__(self) -> int:
    #     return hash((
    #         self.message_type,
    #         self.note,
    #         self.velocity,
    #         self.control,
    #         self.numerator,
    #         self.denominator,
    #         self.key,
    #         self.time,
    #         self.program
    #     ))

    def __repr__(self) -> str:
        representation = f"[{self.message_type.value}]:"

        if self.time is not None:
            representation += f" time={self.time}"

        if self.note is not None:
            representation += f" note={self.note}"

        if self.velocity is not None:
            representation += f" velocity={self.velocity}"

        if self.control is not None:
            representation += f" control={self.control}"

        if self.program is not None:
            representation += f" program={self.program}"

        if self.numerator is not None:
            representation += f" numerator={self.numerator}"

        if self.denominator is not None:
            representation += f" denominator={self.denominator}"

        if self.key is not None:
            representation += f" key={self.key.value}"

        return representation

    @staticmethod
    def from_dict(dictionary: dict) -> Message:
        msg = Message(message_type=MessageType[dictionary.get("message_type", None)], note=dictionary.get("note", None),
                      velocity=dictionary.get("velocity", None), control=dictionary.get("control", None),
                      program=dictionary.get("program", None), numerator=dictionary.get("numerator", None),
                      denominator=dictionary.get("denominator", None), key=dictionary.get("key", None),
                      time=dictionary.get("time", None))

        return msg



from __future__ import annotations

from abc import ABC, abstractmethod

from scoda.elements.message import Message


class AbstractSequence(ABC):
    """Class representing an abstract musical sequence.

    """

    def __init__(self) -> None:
        super().__init__()
        self.messages = []

    @abstractmethod
    def add_message(self, msg: Message) -> None:
        """Adds a message to the sequence.

        For an `AbsoluteSequence` the message has to contain an entry for `time`, according to which it will
        be sorted into the sequence. This is not needed for a `RelativeSequence`.

        Args:
            msg: The message to append

        """
        pass



import json
from pathlib import Path

# General settings

"""Parts per quarter note, amount of ticks used for the representation of a quarter note"""
PPQN: int
"""Maximum velocity value, velocity values usually range from 0 to 127"""
VELOCITY_MAX: int
"""Amount of velocity bins to use, consolidating velocity values into `VELOCITY_BINS` bins"""
VELOCITY_BINS: int
"""MIDI value of lowest note"""
NOTE_LOWER_BOUND: int
"""MIDI value of highest note"""
NOTE_UPPER_BOUND: int
"""Maximum power of 2 defining lower length of notes, e.g., 8 corresponds to PPQN / 2^3, or thirty-second notes"""
NOTE_VALUE_LOWER_BOUND: int
"""Maximum power of 2 defining upper length of notes, e.g., 8 corresponds to PPQN * 2^3, or double whole notes"""
NOTE_VALUE_UPPER_BOUND: int
"""Amount of dots for dotted notes to consider"""
DOTTED_ITERATIONS: int
"""Considered tuplets, e.g., a value of (3, 2) allows for triplets"""
VALID_TUPLETS: list[tuple[int, int]]

# Pattern Recognition

"""Minimum length of patterns to be considered"""
PATTERN_LENGTH_MIN: int
"""Amount of seconds after which greedy pattern matching is applied"""
PATTERN_SECONDS_SEARCH_DURATION: int
"""Regex used to find patterns"""
REGEX_PATTERN: str
"""Regex used to find subpatterns"""
REGEX_SUBPATTERN: str

# Difficulty Parameters

DIFF_DUAL_NOTE_AMOUNT_UPPER_BOUND: float
DIFF_DUAL_NOTE_AMOUNT_LOWER_BOUND: float
DIFF_DUAL_NOTE_CLASSES_UPPER_BOUND: float
DIFF_DUAL_NOTE_CLASSES_LOWER_BOUND: float
DIFF_DUAL_NOTE_CONCURRENT_UPPER_BOUND: float
DIFF_DUAL_NOTE_CONCURRENT_LOWER_BOUND: float
DIFF_DUAL_NOTE_VALUES_UPPER_BOUND: float
DIFF_DUAL_NOTE_VALUES_LOWER_BOUND: float
DIFF_DUAL_DISTANCES_UPPER_BOUND: float
DIFF_DUAL_DISTANCES_LOWER_BOUND: float
DIFF_DUAL_PATTERN_COVERAGE_UPPER_BOUND: float
DIFF_DUAL_PATTERN_COVERAGE_LOWER_BOUND: float
DIFF_DUAL_ACCIDENTALS_UPPER_BOUND: float
DIFF_DUAL_ACCIDENTALS_LOWER_BOUND: float

# Polynomials

"""Coefficients for cubic curve"""
SCALE_CUBIC: list[float]
"""Coefficients for loglike curve"""
SCALE_LOGLIKE: list[float]


def load_from_file():
    settings_file_path = Path(__file__).parent.parent.joinpath("config/settings.json")
    settings_file = open(settings_file_path)
    settings = json.load(settings_file)

    global PPQN
    global VELOCITY_MAX
    global VELOCITY_BINS
    global NOTE_LOWER_BOUND
    global NOTE_UPPER_BOUND
    global NOTE_VALUE_LOWER_BOUND
    global NOTE_VALUE_UPPER_BOUND
    global DOTTED_ITERATIONS
    global VALID_TUPLETS

    global PATTERN_LENGTH_MIN
    global PATTERN_SECONDS_SEARCH_DURATION
    global REGEX_PATTERN
    global REGEX_SUBPATTERN

    global DIFF_DUAL_NOTE_AMOUNT_UPPER_BOUND
    global DIFF_DUAL_NOTE_AMOUNT_LOWER_BOUND
    global DIFF_DUAL_NOTE_CLASSES_UPPER_BOUND
    global DIFF_DUAL_NOTE_CLASSES_LOWER_BOUND
    global DIFF_DUAL_NOTE_CONCURRENT_UPPER_BOUND
    global DIFF_DUAL_NOTE_CONCURRENT_LOWER_BOUND
    global DIFF_DUAL_NOTE_VALUES_LOWER_BOUND
    global DIFF_DUAL_NOTE_VALUES_UPPER_BOUND
    global DIFF_DUAL_DISTANCES_UPPER_BOUND
    global DIFF_DUAL_DISTANCES_LOWER_BOUND
    global DIFF_DUAL_PATTERN_COVERAGE_UPPER_BOUND
    global DIFF_DUAL_PATTERN_COVERAGE_LOWER_BOUND
    global DIFF_DUAL_ACCIDENTALS_UPPER_BOUND
    global DIFF_DUAL_ACCIDENTALS_LOWER_BOUND

    global SCALE_CUBIC
    global SCALE_LOGLIKE

    PPQN = settings["general_settings"]["ppqn"]
    VELOCITY_MAX = settings["general_settings"]["velocity_max"]
    VELOCITY_BINS = settings["general_settings"]["velocity_bins"]
    NOTE_LOWER_BOUND = settings["general_settings"]["note_lower_bound"]
    NOTE_UPPER_BOUND = settings["general_settings"]["note_upper_bound"]
    NOTE_VALUE_LOWER_BOUND = settings["general_settings"]["note_value_lower_bound"]
    NOTE_VALUE_UPPER_BOUND = settings["general_settings"]["note_value_upper_bound"]
    DOTTED_ITERATIONS = settings["general_settings"]["dotted_iterations"]

    VALID_TUPLETS = []
    for tuplet in settings["general_settings"]["tuplets_valid"]:
        VALID_TUPLETS.append(tuple(tuplet))

    PATTERN_LENGTH_MIN = settings["pattern_recognition"]["pattern_length_min"]
    PATTERN_SECONDS_SEARCH_DURATION = settings["pattern_recognition"]["pattern_seconds_search_duration"]
    REGEX_PATTERN = settings["pattern_recognition"]["regex_pattern"]
    REGEX_SUBPATTERN = settings["pattern_recognition"]["regex_subpattern"]

    DIFF_DUAL_NOTE_AMOUNT_UPPER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_amount_upper_bound"]
    DIFF_DUAL_NOTE_AMOUNT_LOWER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_amount_lower_bound"]
    DIFF_DUAL_NOTE_CLASSES_UPPER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_classes_upper_bound"]  # 10 / 4
    DIFF_DUAL_NOTE_CLASSES_LOWER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_classes_lower_bound"]  # 4 / 4
    DIFF_DUAL_NOTE_CONCURRENT_UPPER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_concurrent_upper_bound"]
    DIFF_DUAL_NOTE_CONCURRENT_LOWER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_concurrent_lower_bound"]
    DIFF_DUAL_NOTE_VALUES_UPPER_BOUND = PPQN / settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_values_upper_bound"]  # (PPQN / (2 ** 2))
    DIFF_DUAL_NOTE_VALUES_LOWER_BOUND = PPQN / settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_note_values_lower_bound"]  # (PPQN / (2 ** 0) + (PPQN / (2 ** 0))) / 2
    DIFF_DUAL_DISTANCES_UPPER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_distances_upper_bound"]
    DIFF_DUAL_DISTANCES_LOWER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_distances_lower_bound"]
    DIFF_DUAL_PATTERN_COVERAGE_UPPER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_pattern_coverage_upper_bound"]
    DIFF_DUAL_PATTERN_COVERAGE_LOWER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_pattern_coverage_lower_bound"]
    DIFF_DUAL_ACCIDENTALS_UPPER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_accidentals_upper_bound"]
    DIFF_DUAL_ACCIDENTALS_LOWER_BOUND = settings["difficulty_parameters"]["dual_track_difficulty_parameters"][
        "diff_dual_accidentals_lower_bound"]

    # Generated using http://arachnoid.com
    SCALE_CUBIC = settings["difficulty_parameters"]["polynomials"]["scale_cubic"]
    SCALE_LOGLIKE = settings["difficulty_parameters"]["polynomials"]["scale_loglike"]


load_from_file()



from abc import ABC

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.enumerations.tokenisation_flags import TokenisationFlags
from scoda.exceptions.tokenisation_exception import TokenisationException
from scoda.misc.music_theory import CircleOfFifths
from scoda.misc.scoda_logging import get_logger
from scoda.sequences.sequence import Sequence
from scoda.tokenisation.base_tokenisation import BaseTokeniser

LOGGER = get_logger(__name__)


class BaseMidilikeTokeniser(BaseTokeniser, ABC):

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__()

        self.flags[TokenisationFlags.RUNNING_TIME_SIG] = running_time_sig


class StandardMidilikeTokeniser(BaseMidilikeTokeniser):
    """Tokeniser that uses midi-like temporal representation.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [  4 -  27] ... wait
    [ 28 - 115] ... note on
    [116 - 203] ... note off
    [204 - 218] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 219

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True) -> list[int]:
        tokens = []

        for message in sequence.rel.messages:
            msg_type = message.message_type

            if msg_type == MessageType.WAIT:
                self.cur_rest_buffer += message.time

                self.prv_type = MessageType.WAIT
            elif msg_type == MessageType.NOTE_ON:
                msg_note = message.note

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                self.cur_rest_buffer = 0

                tokens.append(msg_note - 21 + 28)

                self.prv_type = MessageType.NOTE_ON
            elif msg_type == MessageType.NOTE_OFF:
                msg_note = message.note

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                self.cur_rest_buffer = 0

                tokens.append(msg_note - 21 + 116)

                self.prv_type = MessageType.NOTE_OFF
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = message.numerator
                msg_denominator = message.denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                    self.cur_rest_buffer = 0

                    tokens.append(numerator - 2 + 204)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator

        if apply_buffer and self.cur_rest_buffer > 0:
            tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
            self.cur_rest_buffer = 0

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()

        for token in tokens:
            if token <= 3:
                pass
            elif 4 <= token <= 27:
                seq.rel.add_message(Message(message_type=MessageType.WAIT, time=token - 3))
            elif 28 <= token <= 115:
                seq.rel.add_message(Message(message_type=MessageType.NOTE_ON, note=token - 28 + 21))
            elif 116 <= token <= 203:
                seq.rel.add_message(Message(message_type=MessageType.NOTE_OFF, note=token - 116 + 21))
            elif 204 <= token <= 218:
                seq.rel.add_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, numerator=token - 204 + 2, denominator=8))
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq


class RelativeMidilikeTokeniser(BaseMidilikeTokeniser):
    """Tokeniser that uses midi-like temporal representation and relative distances.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [  4 -  27] ... wait
    [ 28 - 202] ... note on with relative note shift
    [203 - 376] ... note off with relative note shift
    [377 - 392] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 393

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

        self.reset()

    def reset_previous(self) -> None:
        super().reset_previous()

        # A4 as base note
        self.prv_note = 69

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True) -> list[int]:
        tokens = []

        for message in sequence.rel.messages:
            msg_type = message.message_type

            if msg_type == MessageType.WAIT:
                self.cur_rest_buffer += message.time

                self.prv_type = MessageType.WAIT
            elif msg_type == MessageType.NOTE_ON or msg_type == MessageType.NOTE_OFF:
                msg_note = message.note

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                self.cur_rest_buffer = 0

                # Get distance
                note_distance = msg_note - self.prv_note
                tokens.append(note_distance + 87 + 28 + (0 if msg_type == MessageType.NOTE_ON else 175))

                self.prv_type = msg_type
                self.prv_note = msg_note
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = message.numerator
                msg_denominator = message.denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                    self.cur_rest_buffer = 0

                    tokens.append(numerator - 2 + 377)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator

        if apply_buffer and self.cur_rest_buffer > 0:
            tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
            self.cur_rest_buffer = 0

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        prv_note = 69  # A4 is base note

        for token in tokens:
            if token <= 3:
                pass
            elif 4 <= token <= 27:
                seq.rel.add_message(Message(message_type=MessageType.WAIT, time=token - 3))
            elif 28 <= token <= 376:
                shifter = 0 if 28 <= token <= 202 else 175
                note = prv_note + token - 28 - 87 - shifter
                prv_note = note

                if not (21 <= note <= 108):
                    LOGGER.warning(f"Invalid note: {note}")
                while note < 21:
                    note += 12
                while note > 108:
                    note -= 12

                if 28 <= token <= 202:
                    seq.rel.add_message(Message(message_type=MessageType.NOTE_ON, note=note))
                else:
                    seq.rel.add_message(Message(message_type=MessageType.NOTE_OFF, note=note))
            elif 377 <= token <= 392:
                seq.rel.add_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, numerator=token - 377 + 2, denominator=8))
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq


class CoFMidilikeTokeniser(BaseMidilikeTokeniser):
    """Tokeniser that uses midi-like temporal representation and circle of fifths distances.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [  4 -  27] ... wait
    [ 28 -  44] ... octave shift between notes
    [ 45 -  56] ... note on without octave in distance on the circle of fifths
    [ 57 -  68] ... note off without octave in distance on the circle of fifths
    [ 69 -  84] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 85

    def __init__(self, running_octave: bool, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

        self.flags[TokenisationFlags.RUNNING_OCTAVE] = running_octave

        self.prv_octave = None

        self.reset()

    def reset_previous(self) -> None:
        super().reset_previous()

        # A4 as base note
        self.prv_note = 69
        self.prv_octave = 4

    def tokenise(self, sequence: Sequence, apply_buffer: bool = True) -> list[int]:
        tokens = []

        for message in sequence.rel.messages:
            msg_type = message.message_type

            if msg_type == MessageType.WAIT:
                self.cur_rest_buffer += message.time

                self.prv_type = MessageType.WAIT
            elif msg_type == MessageType.NOTE_ON or msg_type == MessageType.NOTE_OFF:
                msg_note = message.note

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                # Get distances
                octave_tgt = msg_note // 12 - 1
                octave_src = self.prv_note // 12 - 1
                octave_shift = octave_tgt - octave_src
                assert -8 <= octave_shift <= 8

                cof_dist = CircleOfFifths.get_distance(self.prv_note, msg_note)
                assert -5 <= cof_dist <= 6

                tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                self.cur_rest_buffer = 0

                # Insert octave shift (if necessary) and note distance
                if not (self.prv_octave == octave_tgt and self.flags.get(TokenisationFlags.RUNNING_OCTAVE, False)):
                    tokens.append((octave_shift + 8) + 28)
                shifter_on_off = 45 if msg_type == MessageType.NOTE_ON else 57
                tokens.append((cof_dist + 5) + shifter_on_off)

                self.prv_type = msg_type
                self.prv_note = msg_note
                self.prv_octave = octave_tgt
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = message.numerator
                msg_denominator = message.denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
                    self.cur_rest_buffer = 0

                    tokens.append(numerator - 2 + 69)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator

        if apply_buffer and self.cur_rest_buffer > 0:
            tokens.extend(self._general_tokenise_flush_time_buffer(time=self.cur_rest_buffer, index_time_def=4))
            self.cur_rest_buffer = 0

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        prv_note = 69  # A4 is base note
        prv_octave = 4

        for token in tokens:
            if token <= 3:
                pass
            elif 4 <= token <= 27:
                seq.rel.add_message(Message(message_type=MessageType.WAIT, time=token - 3))
            elif 28 <= token <= 44:
                prv_octave += token - 28 - 8
            elif 45 <= token <= 68:
                note_base = CircleOfFifths.from_distance(prv_note, (token - 45) - 5)
                note = note_base + prv_octave * 12 + 12  # Shifts notes to A0

                prv_note = note

                if not (21 <= note <= 108):
                    LOGGER.warning(f"Invalid note: {note}")
                while note < 21:
                    note += 12
                while note > 108:
                    note -= 12

                if 45 <= token <= 56:
                    seq.rel.add_message(Message(message_type=MessageType.NOTE_ON, note=note))
                else:
                    seq.rel.add_message(Message(message_type=MessageType.NOTE_OFF, note=note))
            elif 69 <= token <= 84:
                seq.rel.add_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, numerator=token - 69 + 2, denominator=8))
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq



from __future__ import annotations

import enum


class MessageType(enum.Enum):
    INTERNAL = "internal"
    SEQUENCE_CONTROL = "sequence_control"
    KEY_SIGNATURE = "key_signature"
    TIME_SIGNATURE = "time_signature"
    CONTROL_CHANGE = "control_change"
    PROGRAM_CHANGE = "program_change"
    NOTE_OFF = "note_off"
    NOTE_ON = "note_on"
    WAIT = "wait"

    def __lt__(self, other):
        values = [e for e in MessageType]
        return values.index(self) < values.index(other)



from __future__ import annotations

import mido

from scoda.enumerations.message_type import MessageType
from scoda.midi.midi_message import MidiMessage


class MidiTrack:

    def __init__(self) -> None:
        super().__init__()
        self.name = ""
        self.messages: [MidiMessage] = []

    @staticmethod
    def parse_mido_track(mido_track) -> MidiTrack:
        track = MidiTrack()

        for msg in mido_track:
            track.messages.append(MidiMessage.parse_mido_message(msg))

        return track

    def to_mido_track(self) -> mido.MidiTrack:
        track = mido.MidiTrack()

        if self.name is not None and self.name != "":
            track.name = self.name

        time_buffer = 0

        for msg in self.messages:
            if hasattr(msg, "time") and msg.time is not None:
                time_buffer += msg.time

            if msg.message_type == MessageType.NOTE_ON:
                track.append(
                    mido.Message("note_on", note=msg.note, velocity=msg.velocity if msg.velocity is not None else 127,
                                 time=int(time_buffer)))
                time_buffer = 0
            elif msg.message_type == MessageType.NOTE_OFF:
                track.append(mido.Message("note_off", note=msg.note, velocity=0, time=int(time_buffer)))
                time_buffer = 0
            elif msg.message_type == MessageType.WAIT:
                pass
            elif msg.message_type == MessageType.TIME_SIGNATURE:
                track.append(mido.MetaMessage("time_signature", numerator=msg.numerator, denominator=msg.denominator,
                                              time=int(time_buffer)))
                time_buffer = 0
            elif msg.message_type == MessageType.KEY_SIGNATURE:
                track.append(mido.MetaMessage("key_signature", key=msg.key.value, time=int(time_buffer)))
                time_buffer = 0
            elif msg.message_type == MessageType.CONTROL_CHANGE:
                track.append(
                    mido.Message("control_change", channel=0, control=msg.control, value=msg.velocity,
                                 time=int(time_buffer)))
                time_buffer = 0

        return track



from __future__ import annotations

import copy
import math
import re
import time
from statistics import mean
from typing import TYPE_CHECKING

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.exceptions.sequence_exception import SequenceException
from scoda.midi.midi_message import MidiMessage
from scoda.midi.midi_track import MidiTrack
from scoda.misc.music_theory import Note, Key, MusicMapping
from scoda.misc.scoda_logging import get_logger
from scoda.misc.util import minmax, simple_regression
from scoda.sequences.abstract_sequence import AbstractSequence
from scoda.settings.settings import NOTE_LOWER_BOUND, NOTE_UPPER_BOUND, PPQN, DIFF_DUAL_DISTANCES_UPPER_BOUND, \
    DIFF_DUAL_DISTANCES_LOWER_BOUND, DIFF_DUAL_PATTERN_COVERAGE_UPPER_BOUND, DIFF_DUAL_PATTERN_COVERAGE_LOWER_BOUND, \
    PATTERN_LENGTH_MIN, REGEX_PATTERN, REGEX_SUBPATTERN, DIFF_DUAL_NOTE_CLASSES_UPPER_BOUND, \
    DIFF_DUAL_NOTE_CLASSES_LOWER_BOUND, \
    DIFF_DUAL_NOTE_AMOUNT_UPPER_BOUND, DIFF_DUAL_NOTE_AMOUNT_LOWER_BOUND, PATTERN_SECONDS_SEARCH_DURATION, \
    DIFF_DUAL_NOTE_CONCURRENT_UPPER_BOUND, DIFF_DUAL_NOTE_CONCURRENT_LOWER_BOUND, DIFF_DUAL_ACCIDENTALS_UPPER_BOUND, \
    DIFF_DUAL_ACCIDENTALS_LOWER_BOUND

if TYPE_CHECKING:
    from scoda.sequences.absolute_sequence import AbsoluteSequence


class RelativeSequence(AbstractSequence):
    """Class representing a sequence with relative message timings.
    """

    LOGGER = get_logger(__name__)

    # General Methods

    def __init__(self) -> None:
        super().__init__()

    def __copy__(self):
        cpy = RelativeSequence()

        for message in self.messages:
            cpy.messages.append(copy.copy(message))

        return cpy

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, RelativeSequence):
            return False

        return self.to_absolute_sequence() == o.to_absolute_sequence()

    # Basic Methods

    def add_message(self, msg: Message) -> None:
        """Adds the given message to the sequence."""
        self.messages.append(msg)

    def concatenate(self, sequences: list[RelativeSequence]) -> None:
        """Concatenates the sequence with the given sequences, resulting in this sequence containing the combined
        messages of itself and the given sequences.

        Args:
            sequences: A sequence which should be appended to this sequence.

        """
        for seq in sequences:
            self.messages.extend(seq.messages)

    def normalise_relative(self) -> None:
        """Removes invalid open and close messages. Consolidates wait messages. Removes double time signatures and key signatures.

        """
        open_messages = dict()
        messages_normalized = []
        wait_buffer = 0

        current_ts_numerator = None
        current_ts_denominator = None
        current_key = None

        for msg in self.messages:
            if msg.message_type == MessageType.WAIT:
                wait_buffer += msg.time
            else:
                if msg.message_type == MessageType.NOTE_ON:
                    note_list = open_messages.get(msg.note, [])
                    note_list.append(msg)
                    open_messages[msg.note] = note_list

                    # Skip message if note is already open
                    if len(note_list) != 1:
                        continue
                elif msg.message_type == MessageType.NOTE_OFF:
                    note_list = open_messages.get(msg.note, [])
                    if len(note_list) > 0:
                        note_list.pop(-1)
                    open_messages[msg.note] = note_list

                    # Skip message if note not yet closed
                    if len(note_list) != 0:
                        continue
                # Remove double time signatures
                elif msg.message_type == MessageType.TIME_SIGNATURE:
                    if msg.numerator != current_ts_numerator or msg.denominator != current_ts_denominator:
                        current_ts_numerator = msg.numerator
                        current_ts_denominator = msg.denominator
                    else:
                        continue
                elif msg.message_type == MessageType.KEY_SIGNATURE:
                    if msg.key != current_key:
                        current_key = msg.key
                    else:
                        continue

                # Insert consolidated wait message
                if wait_buffer > 0:
                    messages_normalized.append(Message(message_type=MessageType.WAIT, time=wait_buffer))
                    wait_buffer = 0

                messages_normalized.append(msg)

        # Repeat procedure for wait messages that occur at the end of the sequence
        if wait_buffer > 0:
            messages_normalized.append(Message(message_type=MessageType.WAIT, time=wait_buffer))

        for key in open_messages.keys():
            note_list = open_messages.get(key, [])

            if len(note_list) > 0:
                for msg in note_list:
                    if msg in messages_normalized:
                        messages_normalized.remove(msg)

        self.messages = messages_normalized

    def pad(self, padding_length) -> None:
        """Pads the sequence to a minimum fixed length.

        Args:
            padding_length: The minimum length this sequence should have after this operation.

        """
        current_length = 0

        for msg in self.messages:
            if msg.message_type == MessageType.WAIT:
                current_length += msg.time

                if current_length >= padding_length:
                    break

        if current_length < padding_length:
            self.messages.append(Message(message_type=MessageType.WAIT, time=padding_length - current_length))

    def split(self, capacities: list[int]) -> list[RelativeSequence]:
        """Splits the sequence into parts of the given capacity.

        Creates up to `len(capacities) + 1` new `RelativeSequence`s, where the first `len(capacities)` entries contain
        sequences of the given capacities, while the last one contains any remaining notes. Messages at the boundaries
        of a capacity are split up and possibly reinserted at the beginning of the next sequence.

        Args:
            capacities: A list of capacities to split the sequence into.

        Returns: A list of `RelativeSequence`s of the desired size.

        """
        split_sequences = []
        working_memory = copy.copy(self.messages)

        current_sequence = RelativeSequence()
        open_messages = dict()

        # Try to split current sequence at given point
        for capacity in capacities:
            next_sequence = RelativeSequence()
            next_sequence_queue = []
            remaining_capacity = capacity

            while remaining_capacity >= 0:
                # Check if end-of-sequence was reached prematurely
                if len(working_memory) == 0:
                    if len(current_sequence.messages) > 0:
                        split_sequences.append(current_sequence)
                        current_sequence = next_sequence
                    break

                # Retrieve next message
                msg = working_memory.pop(0)

                # Check messages, if capacity 0 add to next sequence for most of them
                if msg.message_type == MessageType.NOTE_ON:
                    if remaining_capacity > 0:
                        current_sequence.add_message(msg)
                        open_messages[msg.note] = copy.copy(msg)
                    else:
                        next_sequence_queue.append(msg)
                # For stop messages, add them to the current sequence
                elif msg.message_type == MessageType.NOTE_OFF:
                    current_sequence.add_message(msg)
                    open_messages.pop(msg.note, None)
                elif msg.message_type == MessageType.WAIT:
                    # Can add message in entirety
                    if msg.time <= remaining_capacity:
                        remaining_capacity -= msg.time
                        current_sequence.add_message(msg)
                    # Have to split message
                    else:
                        carry_time = msg.time - remaining_capacity

                        if remaining_capacity > 0:
                            current_sequence.add_message(
                                Message(message_type=MessageType.WAIT, time=remaining_capacity))

                        for key, value in open_messages.items():
                            current_sequence.add_message(Message(message_type=MessageType.NOTE_OFF, note=value.note))
                            next_sequence_queue.append(
                                Message(message_type=MessageType.NOTE_ON, note=value.note, velocity=value.velocity))

                        next_sequence_queue.append(Message(message_type=MessageType.WAIT, time=carry_time))

                        if len(current_sequence.messages) > 0:
                            split_sequences.append(current_sequence)
                        working_memory[0:0] = next_sequence_queue
                        current_sequence = next_sequence
                        break
                else:
                    if remaining_capacity > 0:
                        current_sequence.add_message(msg)
                    else:
                        next_sequence_queue.append(msg)

        # Check if still capacity left
        if len(working_memory) > 0:
            current_sequence.messages.extend(working_memory)

        # Add current sequence if it is not empty
        if len(current_sequence.messages) > 0:
            split_sequences.append(current_sequence)

        return split_sequences

    def scale(self, factor, meta_sequence=None) -> None:
        """Stretches the sequence by the given factor.

        Args:
            factor: Factor to stretch by.
            meta_sequence: Sequence containing the time signatures to apply.

        """
        if factor > 1:
            if not (factor * 1.0).is_integer():
                raise SequenceException("Factor results in non-integer scaling")
        else:
            if not (1 / factor).is_integer():
                raise SequenceException("Factor results in non-integer scaling")

        # Normal case, simply multiply duration
        if factor == 1:
            return
        if factor > 1:
            for msg in self.messages:
                if msg.message_type == MessageType.WAIT:
                    msg.time = msg.time * factor
        # Handle special case, have to consider time signatures
        else:
            from scoda.sequences.sequence import Sequence

            modified_messages = []
            sequence = Sequence(relative_sequence=self)

            if meta_sequence is None:
                meta_sequence = sequence

            amount_consecutive_bars = 1 / factor
            bars = Sequence.sequences_split_bars(
                [sequence, meta_sequence], meta_track_index=1, quantise_note_lengths=False)[0]

            bar_index = 0

            # Handle each bar
            while bar_index < len(bars):
                current_bar = bars[bar_index]
                consecutive_bars = [current_bar]

                # Get all consecutive bars in chunk
                for i in range(1, int(amount_consecutive_bars)):
                    if bar_index + i < len(bars):
                        consecutive_bars.append(bars[bar_index + i])

                # Check if all have same time signature
                if all(cbar.time_signature_numerator == current_bar.time_signature_numerator and
                       cbar.time_signature_denominator == current_bar.time_signature_denominator
                       for cbar in consecutive_bars):
                    for msg in [msg for cbar in consecutive_bars for msg in cbar.sequence.rel.messages]:
                        if msg.message_type == MessageType.WAIT:
                            msg.time = msg.time * factor

                        modified_messages.append(msg)

                    bar_index += len(consecutive_bars)
                # Not all have same time signature
                else:
                    for msg in current_bar.sequence.rel.messages:
                        if msg.message_type == MessageType.WAIT:
                            msg.time = msg.time * factor
                        elif msg.message_type == MessageType.TIME_SIGNATURE:
                            if msg.numerator % (1 / factor) == 0:
                                msg.numerator = int(msg.numerator * factor)
                            else:
                                msg.denominator = int(msg.denominator * (1 / factor))

                        modified_messages.append(msg)

                    bar_index += 1

            self.messages = modified_messages
            self.normalise_relative()

    def transpose(self, transpose_by: int) -> bool:
        """Transposes the sequence by the given amount.

        If the lower or upper bound is undercut over exceeded, these notes are transposed by an octave each.

        Args:
            transpose_by: Half-tone steps to transpose by.

        Returns: `True` if at least one note had to be shifted due to it otherwise being out of bounds.

        """
        had_to_shift = False

        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON or msg.message_type == MessageType.NOTE_OFF:
                msg.note += transpose_by
                while msg.note < NOTE_LOWER_BOUND:
                    had_to_shift = True
                    msg.note += 12
                while msg.note > NOTE_UPPER_BOUND:
                    had_to_shift = True
                    msg.note -= 12
            elif msg.message_type == MessageType.KEY_SIGNATURE:
                msg.key = Key.transpose_key(msg.key, transpose_by)

        return had_to_shift

    # Misc. Methods

    def is_empty(self) -> bool:
        """Checks if the sequence is empty, i.e., no notes are opened.

        Returns: `True` if the sequence is empty, `False` otherwise.

        """
        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON:
                return False
        return True

    def get_key_signature_guess(self) -> Key:
        """Determines the best key based on which key induces the minimum amount of additional accidentals.

        Returns: The best-fitting key for this bar.

        """
        for msg in self.messages:
            if msg.message_type == MessageType.KEY_SIGNATURE:
                return msg.key
            if msg.message_type == MessageType.WAIT:
                break

        key_candidates = []
        for _ in MusicMapping.KeyNoteMapping:
            key_candidates.append(0)

        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON:
                for i, (_, key_notes) in enumerate(MusicMapping.KeyNoteMapping.items()):
                    if Note(msg.note % 12) not in key_notes[0]:
                        key_candidates[i] += 1

        best_index = 0
        best_solution = math.inf
        best_solution_accidentals = math.inf
        key_note_mapping = list(MusicMapping.KeyNoteMapping.items())

        for i in range(0, len(key_candidates)):
            if key_candidates[i] <= best_solution:
                if key_candidates[i] < best_solution or key_note_mapping[i][1][1] < best_solution_accidentals:
                    best_index = i
                    best_solution = key_candidates[i]
                    best_solution_accidentals = key_note_mapping[i][1][1]

        guessed_key = [key for key in MusicMapping.KeyNoteMapping][best_index]
        return guessed_key

    def get_sequence_duration_relation(self) -> float:
        """Calculates the duration of the sequence in multiples of the `PPQN`.

        Returns: The duration of the sequence as a multiple of the `PPQN`.

        """
        duration = 0

        for msg in self.messages:
            if msg.message_type == MessageType.WAIT:
                duration += msg.time

        return duration / PPQN

    def to_absolute_sequence(self) -> AbsoluteSequence:
        """Converts this `RelativeSequence` to an `AbsoluteSequence`.

        Returns: The absolute representation of this sequence.

        """
        from scoda.sequences.absolute_sequence import AbsoluteSequence
        absolute_sequence = AbsoluteSequence()
        current_point_in_time = 0
        cap_message_exists = True

        for msg in self.messages:
            if msg.message_type == MessageType.WAIT:
                current_point_in_time += msg.time
                cap_message_exists = False
            else:
                message_to_add = copy.copy(msg)
                message_to_add.time = current_point_in_time
                absolute_sequence._add_message_unsorted(message_to_add)
                cap_message_exists = True

        absolute_sequence.normalise_absolute()

        if not cap_message_exists:
            absolute_sequence.add_message(Message(message_type=MessageType.INTERNAL, time=current_point_in_time))

        return absolute_sequence

    def to_midi_track(self) -> MidiTrack:
        """Converts the sequence to a `MidiTrack`.

        Returns: The corresponding `MidiTrack`.

        """
        track = MidiTrack()

        for msg in self.messages:
            track.messages.append(MidiMessage.parse_internal_message(msg))

        return track

    # Difficulty Methods

    def diff_accidentals(self, key: Key) -> float:
        """Calculates the difficulty of the sequence based on the amount of accidentals needed.

        Args:
            key: A fixed key of the sequence.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        note_mapping = MusicMapping.KeyNoteMapping[key]
        violations = 0

        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON:
                if Note(msg.note % 12) not in note_mapping[0]:
                    violations += 1

        relation = violations / self.get_sequence_duration_relation()
        scaled_relation = simple_regression(DIFF_DUAL_ACCIDENTALS_UPPER_BOUND, 1, DIFF_DUAL_ACCIDENTALS_LOWER_BOUND, 0,
                                            relation)

        return minmax(0, 1, scaled_relation)

    def diff_concurrent_notes(self) -> float:
        """Calculates the difficulty of the sequence based on the amount of concurrent notes.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        concurrent_notes = []
        open_notes = set()
        notes_to_open = set()
        notes_to_close = set()

        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON:
                notes_to_open.add(msg.note)
            elif msg.message_type == MessageType.NOTE_OFF:
                notes_to_close.add(msg.note)
            elif msg.message_type == MessageType.WAIT:
                if len(open_notes) > 0:
                    concurrent_notes.append(len(open_notes))

                if len(notes_to_close) > 0:
                    for note in notes_to_close:
                        open_notes.remove(note)
                    notes_to_close = set()

                if len(notes_to_open) > 0:
                    for note in notes_to_open:
                        open_notes.add(note)
                    notes_to_open = set()

        # Handle end of sequence
        if len(open_notes) > 0:
            concurrent_notes.append(len(open_notes))

        # Handle no entries
        if len(concurrent_notes) == 0:
            concurrent_notes.append(0)

        scaled_difficulty = simple_regression(DIFF_DUAL_NOTE_CONCURRENT_UPPER_BOUND, 1,
                                              DIFF_DUAL_NOTE_CONCURRENT_LOWER_BOUND, 0,
                                              mean(concurrent_notes))

        return minmax(0, 1, scaled_difficulty)

    def diff_distances(self) -> float:
        """Calculates the difficulty of the sequence based on the distances between notes.

        Here, only the top 15% of distances are considered, in order not to disregard notes due to dilution.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        notes_played = []
        current_notes = []
        distances = []

        for msg in self.messages:
            if msg.message_type == MessageType.WAIT and len(current_notes) > 0:
                notes_played.append(sorted(current_notes))
                current_notes = []
            elif msg.message_type == MessageType.NOTE_ON:
                current_notes.append(msg.note)

        for i in range(1, len(notes_played)):
            distance_lower = abs(notes_played[i][0] - notes_played[i - 1][0])
            distance_higher = abs(notes_played[i][-1] - notes_played[i - 1][-1])
            distance = max(distance_lower, distance_higher)
            distances.append(distance)

        # If bar is empty, return easiest difficulty
        if len(distances) == 0:
            return 0

        high_distances_mean = mean(sorted(distances, reverse=True)[0: max(1, math.ceil((len(distances) * 0.15)))])

        scaled_difficulty = simple_regression(DIFF_DUAL_DISTANCES_UPPER_BOUND, 1, DIFF_DUAL_DISTANCES_LOWER_BOUND, 0,
                                              high_distances_mean)

        return minmax(0, 1, scaled_difficulty)

    def diff_key(self, key: Key = None) -> float:
        """Calculates the difficulty of the sequence based on the key it is in.

        Here, a key that is further from C is considered more difficult to play, since the performer has to consider
        more accidentals. Furthermore, if no key is specified, a key is guessed based on the amount of the induced
        accidentals that would have to be played.

        Args:
            key: A fixed key of the sequence, will prevent the program from trying to determine one.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        key_signature = key

        for msg in self.messages:
            if msg.message_type == MessageType.KEY_SIGNATURE:
                if key_signature is not None and key_signature is not msg.key:
                    RelativeSequence.LOGGER.info(f"Key was {key_signature}, now is {msg.key}.")
                    key_signature = None
                    break
                key_signature = msg.key
            if msg.message_type == MessageType.WAIT:
                break

        # Have to guess key signature based on induced accidentals
        if key_signature is None:
            key_signature = self.get_key_signature_guess()

        # Check how many accidentals this key uses
        _, accidentals = MusicMapping.KeyNoteMapping[key_signature]

        scaled_difficulty = simple_regression(7, 1, 0, 0, accidentals)
        return minmax(0, 1, scaled_difficulty)

    def diff_note_amount(self) -> float:
        """Calculates difficulty of the sequence based on the amount of notes played.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        amount_notes_played = 0

        if self.get_sequence_duration_relation() == 0:
            return 0

        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON:
                amount_notes_played += 1

        relation = amount_notes_played / self.get_sequence_duration_relation()

        scaled_difficulty = simple_regression(DIFF_DUAL_NOTE_AMOUNT_UPPER_BOUND, 1, DIFF_DUAL_NOTE_AMOUNT_LOWER_BOUND,
                                              0,
                                              relation)

        return minmax(0, 1, scaled_difficulty)

    def diff_note_classes(self) -> float:
        """Calculates difficulty of the sequence based on the amount of note classes (different notes played) in
        relation to the overall amount of messages.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        note_classes = []

        for msg in self.messages:
            if msg.message_type == MessageType.NOTE_ON and msg.note not in note_classes:
                note_classes.append(msg.note)

        # If sequence is empty, return easiest difficulty
        if len(note_classes) == 0:
            return 0

        relation = len(note_classes) / self.get_sequence_duration_relation()
        scaled_relation = simple_regression(DIFF_DUAL_NOTE_CLASSES_UPPER_BOUND, 1, DIFF_DUAL_NOTE_CLASSES_LOWER_BOUND,
                                            0,
                                            relation)

        return minmax(0, 1, scaled_relation)

    def diff_pattern(self) -> float:
        """Calculates the difficulty of the sequence based on the patterns of the start messages.

        If a sequence contains patterns, i.e., if a building block is reused several times, the sequence is easier to
        play, since the player has to read fewer notes.

        Returns: A value from 0 (low difficulty) to 1 (high difficulty).

        """
        notes_played = []
        current_bin = []

        for msg in self.messages:
            if msg.message_type == MessageType.WAIT and len(current_bin) > 0:
                notes_played.extend(sorted(current_bin, key=lambda message: message.note))
            elif msg.message_type == MessageType.NOTE_ON:
                notes_played.append(msg)

        string_representation = ""

        # Create string representation
        for i in range(1, len(notes_played)):
            value = notes_played[i].note - notes_played[i - 1].note
            if value >= 0:
                string_representation += "+"
            else:
                string_representation += "-"

            string_representation += str(abs(value))

        # Obtain patterns, switch to greedy method if pattern matching takes too long
        try:
            results = RelativeSequence._match_pattern(string_representation, start_time=time.time(),
                                                      max_duration=PATTERN_SECONDS_SEARCH_DURATION)
        except TimeoutError:
            results = RelativeSequence._greedy_match_pattern(string_representation)

        results_with_coverage = []

        # Determine coverage and length of found pattern
        for result in results:
            uncovered = string_representation
            length = 0
            for match in result:
                uncovered = uncovered.replace(match, "")
                length += match.count("+") + match.count("-")
            coverage = (len(string_representation) - len(uncovered)) / len(string_representation)
            results_with_coverage.append((coverage, length, result))

        # Determine best fitting result
        if len(results_with_coverage) > 0:
            best_fit = max(results_with_coverage, key=lambda x: x[0] / x[1])

            bound_difficulty = simple_regression(DIFF_DUAL_PATTERN_COVERAGE_UPPER_BOUND, 1,
                                                 DIFF_DUAL_PATTERN_COVERAGE_LOWER_BOUND, 0,
                                                 best_fit[0] / best_fit[1])

            return minmax(0, 1, bound_difficulty)
        else:
            return 1

    # Static Functions

    @staticmethod
    def _match_pattern(current_representation, start_time, max_duration=10) -> [[str]]:
        """Finds all possible combinations of patterns for input string.

        Recursively finds patterns that fit the input ```current_representation```, removes these patterns from the
        input and tries to match the resulting string. Returns all possible combinations of such matches.

        Args:
            current_representation: The current string to pattern-match.

        Returns: A list of lists, containing the valid patterns that can be matched, in this order, to the input string.

        """
        if time.time() - start_time > max_duration:
            raise TimeoutError

            # Store matches found in this iteration
        local_matches = []
        current_pattern_length = PATTERN_LENGTH_MIN

        # Increase length of pattern each step
        while True:
            matches = re.findall(REGEX_PATTERN.format(p_len=current_pattern_length), current_representation)

            # If no more matches, end calculation
            if len(matches) == 0:
                break

            for match in matches:
                matched_string = match[0]

                # Check if match either already handled, or not a valid pattern (since it contains pattern itself)
                if matched_string not in local_matches and re.match(REGEX_SUBPATTERN, matched_string) is None:
                    local_matches.append(matched_string)

            current_pattern_length += 1

        results = []

        # Handle found matches: Remove pattern from input and try to find patterns in the resulting string
        for local_match in local_matches:
            modified_string = current_representation.replace(local_match, "")

            recursive_results = RelativeSequence._match_pattern(modified_string, start_time, max_duration=max_duration)

            # Consider also only parent match
            if local_match not in results:
                results.append([local_match])

            # Add current results to recursive results
            for recursive_result in recursive_results:
                result = [local_match]
                result.extend(recursive_result)
                results.append(result)

        return results

    @staticmethod
    def _greedy_match_pattern(current_representation) -> [[str]]:
        """Finds possible combinations of patterns for input string, fixing patterns greedily.

        Only considers the first found pattern for further matching, reducing the time needed to pattern match greatly.

        Args:
            current_representation: The current string to pattern-match.

        Returns: A list of lists, containing the valid patterns that can be matched, in this order, to the input string.

        """
        # Store matches found in this iteration
        local_match = None
        current_pattern_length = PATTERN_LENGTH_MIN

        # Increase length of pattern each step
        while True:
            matches = re.findall(REGEX_PATTERN.format(p_len=current_pattern_length), current_representation)
            match = matches[0] if len(matches) > 0 else None

            # If no more matches, end calculation
            if match is None:
                break

            matched_string = match[0]

            # Check if match either already handled, or not a valid pattern (since it contains pattern itself)
            if re.match(REGEX_SUBPATTERN, matched_string) is None:
                local_match = matched_string

            current_pattern_length += 1

        results = []

        # Handle found matches: Remove pattern from input and try to find patterns in the resulting string
        if local_match is not None:
            modified_string = current_representation.replace(local_match, "")

            recursive_results = RelativeSequence._greedy_match_pattern(modified_string)

            # Consider also only parent match
            if local_match not in results:
                results.append([local_match])

            # Add current results to recursive results
            for recursive_result in recursive_results:
                result = [local_match]
                result.extend(recursive_result)
                results.append(result)

        return results



from __future__ import annotations

import copy

from scoda.elements.track import Track
from scoda.sequences.sequence import Sequence


class Composition:
    """Class representing a composition that contains (multiple) tracks.
    """

    def __init__(self, tracks: [Track]) -> None:
        super().__init__()
        self.tracks = tracks

    def __copy__(self) -> Composition:
        tracks = [copy.copy(track) for track in self.tracks]

        cpy = Composition(tracks)
        return cpy

    @staticmethod
    def from_midi_file(file_path: str, track_indices: [[int]],
                       meta_track_indices: [int], meta_track_index: int = 0) -> Composition:
        """Creates a new composition from the given MIDI file.

        Args:
            file_path: Path to the MIDI file
            track_indices: Array of arrays of track indices, indices in the same
                sub-array will be merged to a single track
            meta_track_indices: Indices of tracks to consider for meta messages
            meta_track_index: Which final track to merge the meta messages into

        Returns: The created composition
        """
        # Load sequence from file
        merged_sequences = Sequence.sequences_load(file_path=file_path,
                                                   track_indices=track_indices,
                                                   meta_track_indices=meta_track_indices,
                                                   target_meta_track_index=meta_track_index)

        # Quantisation
        for sequence in merged_sequences:
            sequence.quantise_and_normalise()

        # Load composition from sequence
        return Composition.from_sequences(merged_sequences, meta_track_index)

    @staticmethod
    def from_sequences(sequences, meta_track_index: int = 0) -> Composition:
        # Split sequence into bars
        tracks_bars = Sequence.sequences_split_bars(sequences, meta_track_index=meta_track_index)

        # Create tracks from bars
        tracks = []
        for track_index in range(0, len(tracks_bars)):
            track = Track(tracks_bars[track_index])
            tracks.append(track)

        # Create composition from tracks
        composition = Composition(tracks)
        return composition

    def to_sequences(self):
        sequences = []
        for track in self.tracks:
            sequences.append(track.to_sequence())
        return sequences

    def save(self, file_path: str):
        sequences = self.to_sequences()
        Sequence.sequences_save(sequences, file_path)


import math
from abc import ABC

from scoda.elements.message import Message
from scoda.enumerations.message_type import MessageType
from scoda.enumerations.tokenisation_flags import TokenisationFlags
from scoda.exceptions.tokenisation_exception import TokenisationException
from scoda.sequences.sequence import Sequence
from scoda.tokenisation.base_tokenisation import BaseTokeniser


class BaseGridlikeTokeniser(BaseTokeniser, ABC):
    def __init__(self, running_time_sig: bool) -> None:
        super().__init__()

        self.flags[TokenisationFlags.RUNNING_TIME_SIG] = running_time_sig


class GridlikeTokeniser(BaseGridlikeTokeniser):
    """Tokeniser that uses grid-like temporal representation. Note that input sequences are expected to represent bars,
    as the grid size definition is redone for each input sequence.

    [        0] ... pad
    [        1] ... start
    [        2] ... stop
    [        3] ... bar separator
    [  4 -  27] ... grid time definition
    [ 28 - 115] ... note on
    [116 - 203] ... note off
    [204 - 218] ... time signature numerator in eights from 2/8 to 16/8
    """

    VOCAB_SIZE = 219

    def __init__(self, running_time_sig: bool) -> None:
        super().__init__(running_time_sig)

    def tokenise(self, bar_sequence: Sequence, apply_buffer: bool = True) -> list[int]:
        tokens = []
        min_grid_size = self.set_max_rest_value

        # First pass to get minimum grid size
        for message in bar_sequence.rel.messages:
            msg_type = message.message_type

            if msg_type == MessageType.WAIT:
                self.cur_rest_buffer += message.time
            elif msg_type in [MessageType.NOTE_ON, MessageType.NOTE_OFF, MessageType.TIME_SIGNATURE]:
                if self.cur_rest_buffer > 0:
                    min_grid_size = math.gcd(int(min_grid_size), int(self.cur_rest_buffer))
                self.cur_rest_buffer = 0

        # Calculate grid size for trailing waits
        if self.cur_rest_buffer > 0:
            min_grid_size = math.gcd(int(min_grid_size), int(self.cur_rest_buffer))
        self.cur_rest_buffer = 0

        # Limit grid size
        min_grid_size = min(min_grid_size, self.set_max_rest_value)

        tokens.append(min_grid_size + 3)

        # Second pass to generate tokens
        for message in bar_sequence.rel.messages:
            msg_type = message.message_type

            if msg_type == MessageType.WAIT:
                self.cur_rest_buffer += message.time

                self.prv_type = MessageType.WAIT
            elif msg_type == MessageType.NOTE_ON:
                msg_note = message.note

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                tokens.extend(self._gridlike_tokenise_flush_grid_buffer(min_grid_size=min_grid_size, wait_token=3))
                tokens.append(msg_note - 21 + 28)

                self.prv_type = MessageType.NOTE_ON
            elif msg_type == MessageType.NOTE_OFF:
                msg_note = message.note

                if not (21 <= msg_note <= 108):
                    raise TokenisationException(f"Invalid note: {msg_note}")

                tokens.extend(self._gridlike_tokenise_flush_grid_buffer(min_grid_size=min_grid_size, wait_token=3))
                tokens.append(msg_note - 21 + 116)

                self.prv_type = MessageType.NOTE_OFF
            elif msg_type == MessageType.TIME_SIGNATURE:
                msg_numerator = message.numerator
                msg_denominator = message.denominator

                numerator = self._time_signature_to_eights(msg_numerator, msg_denominator)

                if not (self.prv_numerator == numerator and self.flags.get(TokenisationFlags.RUNNING_TIME_SIG, False)):
                    tokens.extend(self._gridlike_tokenise_flush_grid_buffer(min_grid_size=min_grid_size, wait_token=3))
                    tokens.append(numerator - 2 + 204)

                self.prv_type = MessageType.TIME_SIGNATURE
                self.prv_numerator = numerator

        if apply_buffer and self.cur_rest_buffer > 0:
            tokens.extend(self._gridlike_tokenise_flush_grid_buffer(min_grid_size=min_grid_size, wait_token=3))

        return tokens

    @staticmethod
    def detokenise(tokens: list[int]) -> Sequence:
        seq = Sequence()
        prv_type = None
        min_grid_size = math.nan

        for token in tokens:
            if token <= 2:
                prv_type = "sequence_control"
            elif token == 3:
                if math.isnan(min_grid_size):
                    raise TokenisationException(f"Grid size not initialised")

                seq.add_relative_message(Message(MessageType.WAIT, time=min_grid_size))

                prv_type = MessageType.INTERNAL
            elif 4 <= token <= 27:
                if prv_type == MessageType.WAIT:
                    raise TokenisationException(f"Illegal consecutive grid size definition")

                min_grid_size = token - 3

                prv_type = MessageType.WAIT
            elif 28 <= token <= 115:
                seq.rel.add_message(Message(message_type=MessageType.NOTE_ON, note=token - 28 + 21))

                prv_type = MessageType.NOTE_ON
            elif 116 <= token <= 203:
                seq.rel.add_message(Message(message_type=MessageType.NOTE_OFF, note=token - 116 + 21))

                prv_type = MessageType.NOTE_OFF
            elif 204 <= token <= 218:
                seq.rel.add_message(
                    Message(message_type=MessageType.TIME_SIGNATURE, numerator=token - 204 + 2, denominator=8))

                prv_type = MessageType.TIME_SIGNATURE
            else:
                raise TokenisationException(f"Encountered invalid token during detokenisation: {token}")

        return seq
